#coding:utf-8
#
# PROGRAM/MODULE: saturnin
# FILE:           saturnin/lib/command.py
# DESCRIPTION:    Base classes for Saturnin CLI commands
# CREATED:        19.1.2021
#
# The contents of this file are subject to the MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Copyright (c) 2021 Firebird Project (www.firebirdsql.org)
# All Rights Reserved.
#
# Contributor(s): Pavel Císař (original code)
#                 ______________________________________

"""Saturnin base classes for CLI commands


"""

from __future__ import annotations
from typing import Dict
from weakref import proxy
from argparse import ArgumentParser, Namespace
from traceback import format_exception_only
from pkg_resources import iter_entry_points
from firebird.base.signal import eventsocket
from saturnin.base import Error

class Command:
    """Base class for Saturnin CLI commands.
    """
    def __init__(self, name: str, description: str):
        self.name: str = name
        self.description: str = description
    def print_table(self, headers: List[str], data: List[List[str]]) -> None:
        """Print table.
        """
        sizes = [len(i) for i in headers]
        for row in data:
            for i, col in zip(range(len(sizes)), row):
                sizes[i] = max(sizes[i], len(col))
        fmt = ' '.join(f'{{!s:{size}}}' for size in sizes)
        fmt += '\n'
        self.out(fmt.format(*headers))
        self.out(fmt.format(*['-' * size for size in sizes]))
        for row in data:
            self.out(fmt.format(*row))
    def set_arguments(self, manager: CommandManager, parser: ArgumentParser) -> None:
        """Set CLI arguments.
        """
    def on_error(self, exc: Exception):
        """Called when exception is raised in `.run()`.
        """
    def run(self, args: Namespace) -> None:
        """Command execution.

        Arguments:
            args: Collected argument values.
        """
    @eventsocket
    def out(self, msg: str, end='') -> None:
        """Print informational message.
        """


class HelpCommand(Command):
    """Show help for commands.
    """
    def __init__(self):
        super().__init__('help', "Show help for commands.")
        self.parser: ArgumentParser = None
    def set_arguments(self, manager: CommandManager, parser: ArgumentParser) -> None:
        """Set command arguments.
        """
        self.manager = proxy(manager)
        parser.add_argument('command')
    def run(self, args: Namespace) -> None:
        """Command execution.

        Arguments:
            args: Collected argument values.
        """
        self.manager.parser.parse_args([args.command, '--help'])

class CommandManager:
    """Saturni CLI command manager.
    """
    def __init__(self, parser: ArgumentParser):
        """
        Arguments:
            parser: Argument parser.
        """
        self.commands: Dict[str, Command]= {}
        self.parser: ArgumentParser = parser
        self.subparsers = \
            self.parser.add_subparsers(metavar='', prog=f'\n  {parser.prog}', required=True)
        self.parser._subparsers.title = 'Commands'
    def register_command(self, cmd: Command) -> None:
        """Registers command.

        Arguments:
            cmd: Command to be registered.

        Raises:
            Error: If command is already registered.
        """
        if cmd.name in self.commands:
            raise Error(f"Command {cmd.name} already registered")
        self.commands[cmd.name] = cmd
        cmd_parser = self.subparsers.add_parser(cmd.name, description=cmd.description,
                                                help=cmd.description)
        cmd.set_arguments(self, cmd_parser)
        cmd.out = self.out
        cmd_parser.set_defaults(runner=cmd.run)
        cmd_parser.set_defaults(err_handler=cmd.on_error)
    def out(self, msg: str, end='') -> None:
        """Print message to stdout.
        """
        print(msg, end=end, flush=True)
    def load_commands(self, group: str) -> List[Command]:
        """Load commands registered in `options.entry_point` section of `setup.cfg` file.

        Arguments:
            group: Entry-point group name.

        Example:
            ::

               # setup.cfg:

               [options.entry_points]
               saturnin.commands =
                   saturnin.config = saturnin.platform.commands.config:ConfigCommand
                   saturnin.runner = saturnin.platform.commands.run:RunCommand

               # will be loaded with:

               load_commands('saturnin.commands')
        """
        for cmd in (entry.load() for entry in iter_entry_points(group)):
            self.register_command(cmd())
    def run(self, args=None) -> None:
        """Process arguments and execute selected command.
        """
        args = self.parser.parse_args()
        try:
            args.runner(args)
        except Exception as exc:
            try:
                args.err_handler(exc)
            except:
                pass
            msg = ''.join(format_exception_only(type(exc), exc))
            _, msg = msg.split(': ', 1)
            self.parser.exit(1, msg)




#coding:utf-8
#
# PROGRAM/MODULE: saturnin
# FILE:           saturnin/component/bundle.py
# DESCRIPTION:    Service budle controller
# CREATED:        5.12.2020
#
# The contents of this file are subject to the MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Copyright (c) 2020 Firebird Project (www.firebirdsql.org)
# All Rights Reserved.
#
# Contributor(s): Pavel Císař (original code)
#                 ______________________________________

"""Saturnin service budle controller


"""

from __future__ import annotations
from typing import List
import os
import uuid
import weakref
import warnings
import zmq
from configparser import ConfigParser, ExtendedInterpolation, DEFAULTSECT
from firebird.base.types import ZMQDomain
from firebird.base.collections import Registry
from firebird.base.config import ConfigListOption
from firebird.base.logging import LoggingIdMixin, get_logger
from firebird.base.trace import TracedMixin
from saturnin.base import Error, Config, ChannelManager
from .controller import ThreadController, ServiceExecConfig
from .registry import get_service_desciptors

#: Configuration section name for local service addresses
SECTION_LOCAL_ADDRESS = 'local_address'
#: Configuration section name for node service addresses
SECTION_NODE_ADDRESS = 'node_address'
#: Configuration section name for network service addresses
SECTION_NET_ADDRESS = 'net_address'
#: Configuration section name for service UIDs
SECTION_SERVICE_UID = 'service_uid'
#: Configuration section name for service peer UIDs
SECTION_PEER_UID = 'peer_uid'
#: Configuration section name for service bundle
SECTION_BUNDLE = 'bundle'

class ServiceBundleConfig(Config):
    """Service bundle configuration.
    """
    def __init__(self, name: str):
        super().__init__(name)
        self.agents: ConfigListOption = ConfigListOption('agents', "Agent UIDs",
                                                         ServiceExecConfig, required=True)

class BundleThreadController(LoggingIdMixin, TracedMixin):
    """Service controller that manages collection of services executed in separate threads.
    """
    def __init__(self, *, parser: ConfigParser=None, manager: ChannelManager=None):
        """
        Arguments:
            parser: ConfigParser instance to be used for bundle configuration.
            manager: ChannelManager to be used.
        """
        self.log_context = None
        self.mngr: ChannelManager = manager
        self._ext_mngr: bool = manager is not None
        if manager is None:
            self.mngr = ChannelManager(zmq.Context.instance())
            self.mngr.log_context = weakref.proxy(self)
        #: ConfigParser with service bundle configuration
        self.config: ConfigParser = \
            ConfigParser(interpolation=ExtendedInterpolation()) if parser is None else parser
        #: List with ThreadControllers for all service instances in bundle
        self.services: List[ThreadController] = []
        #: Registry with ServiceDescriptors for services that could be run
        self.available_services: Registry = Registry(get_service_desciptors())
        #
        self.config[SECTION_LOCAL_ADDRESS] = {}
        self.config[SECTION_NODE_ADDRESS] = {}
        self.config[SECTION_NET_ADDRESS] = {}
        self.config[SECTION_SERVICE_UID] = {}
        self.config[SECTION_PEER_UID] = {}
        # Defaults
        self.config[DEFAULTSECT]['here'] = os.getcwd()
        # Assign Agent IDs for available services
        self.config[SECTION_SERVICE_UID].update((sd.agent.name, sd.agent.uid.hex) for sd
                                                in self.available_services)
        #
    def configure(self, *, section: str=SECTION_BUNDLE) -> None:
        """
        Arguments:
            section: Configuration section with bundle definition.
        """
        svc_cfg: ServiceExecConfig = ServiceExecConfig(section)
        bundle_cfg: ServiceBundleConfig = ServiceBundleConfig(section)
        bundle_cfg.load_config(self.config)
        bundle_cfg.validate()
        # Assign Peer IDs to service sections (instances)
        peer_uids = {section.name: uuid.uuid1() for section in bundle_cfg.agents.value}
        self.config[SECTION_PEER_UID].update((k, v.hex) for k, v in peer_uids.items())
        #
        #
        for svc_cfg in bundle_cfg.agents.value:
            svc_cfg.validate()
            if svc_cfg.agent.value in self.available_services:
                controller = ThreadController(self.available_services[svc_cfg.agent.value],
                                              name=svc_cfg.name, peer_uid=peer_uids[svc_cfg.name],
                                              manager=self.mngr)
                self.services.append(controller)
            else:
                self.services.clear()
                raise Error(f"Unknonw agent in section '{svc_cfg.name}'")
    def start(self, *, timeout: int=10000) -> None:
        """Start all services in bundle.

        Arguments:
            timeout: Timeout for starting each service. None (infinity), or a floating
                     point number specifying a timeout for the operation in seconds (or
                     fractions thereof) [Default: 10s].

        Important:
            Services are started in order they are listed in bundle configuration.
            If any service fails to start, all previously started services are stopped.

        Raises:
            ServiceError: On error in communication with service.
            TimeoutError: When service does not start in time.
        """
        for controller in self.services:
            try:
                controller.configure(self.config, controller.name)
                controller.log_context = self.log_context
                controller.start(timeout=timeout)
                if controller.endpoints:
                    # Update addresses for binded endpoints
                    for name, addresses in controller.endpoints.items():
                        opt_name = f'{controller.name}.{name}'
                        for address in addresses:
                            if address.domain == ZMQDomain.LOCAL:
                                self.config[SECTION_LOCAL_ADDRESS][opt_name] = address
                            elif address.domain == ZMQDomain.NODE:
                                self.config[SECTION_NODE_ADDRESS][opt_name] = address
                            else:
                                self.config[SECTION_NET_ADDRESS][opt_name] = address
            except:
                self.stop()
                raise
    def stop(self, *, timeout: int=10000) -> None:
        """Stop all runing services in bundle. The services are stopped in the reverse
        order in which they were started.

        Arguments:
            timeout: Timeout for stopping each service. None (infinity), or a floating
                     point number specifying a timeout for the operation in seconds (or
                     fractions thereof) [Default: 10s].

        Raises:
            ServiceError: On error in communication with service.
            TimeoutError: When service does not stop in time.
        """
        for controller in reversed(self.services):
            try:
                controller.stop(timeout=timeout)
            except Exception as exc:
                get_logger(self).error("Error while stopping the service: {args[0]}", exc)
                if controller.is_running():
                    warnings.warn(f"Stopping service {controller.name} failed, service thread terminated", RuntimeWarning)
                    controller.terminate()
    def join(self, timeout=None) -> None:
        """Wait until all services stop.

        Arguments:
            timeout: Floating point number specifying a timeout for the operation in
                     seconds (or fractions thereof).
        """
        for svc in self.services:
            svc.join(timeout)
